from datetime import (
    date,
    datetime,
    time,
)
from typing import (
    TYPE_CHECKING,
    Iterable,
    List,
    Optional,
    Tuple,
)

from django.apps import (
    apps,
)
from django.db import (
    connection,
)

from m3_db_utils.models import (
    ModelEnumValue,
)

from edu_rdm_integration import (
    consts,
)
from edu_rdm_integration.consts import (
    DATE_FORMAT,
)
from edu_rdm_integration.models import (
    RegionalDataMartModelEnum,
)
from edu_rdm_integration.utils import (
    split_by_academic_years,
    split_interval_by_delta,
)


if TYPE_CHECKING:
    from educommon.audit_log.models import (
        AuditLog,
    )


class BaseEduLogGenerator:
    """
    Базовый класс генератора логов для указанной модели РВД за определенный период времени.

    Для каждой модели РВД есть модели в продукте, создание экземпляров которых, сигнализирует о необходимости сбора
    и выгрузки данных в РВД. Модели можно найти в
    edu_rdm_integration/mapping.py MODEL_FIELDS_LOG_FILTER, принадлежность к конкретной
    модели РВД необходимо определять в функциях.
    """

    def generate(
        self,
        model: ModelEnumValue,
        logs_period_started_at: datetime = datetime.combine(date.today(), time.min),
        logs_period_ended_at: datetime = datetime.combine(date.today(), time.max),
        school_ids: Optional[Tuple[int]] = (),
    ) -> List['AuditLog']:
        """
        Возвращает список сгенерированных экземпляров модели AuditLog.

        Формирование логов производится для указанной модели РВД за указанный период времени.

        Args:
            model: значение модели РВД из модели-перечисления;
            logs_period_started_at: начало периода формирования логов;
            logs_period_ended_at: конец периода формирования логов;
            school_ids: список идентификаторов школ.
        """
        generate_logs_method = getattr(self, f'_generate_{model.key.lower()}_logs')

        logs = generate_logs_method(
            logs_period_started_at=logs_period_started_at,
            logs_period_ended_at=logs_period_ended_at,
            school_ids=school_ids,
        )

        return logs


class BaseFirstCollectModelsDataCommandsGenerator:
    """Класс, который генерирует список данных для формирования команд для сбора данных РВД."""

    def __init__(
        self,
        models: Iterable[str],
        logs_period_started_at=datetime.combine(date.today(), time.min),
        logs_period_ended_at=datetime.combine(date.today(), time.min),
        logs_sub_period_days=consts.LOGS_SUB_PERIOD_DAYS,
        batch_size=consts.BATCH_SIZE,
    ):
        """Инициализация."""
        # Если модели не указаны, берется значение по умолчанию - все модели:
        models = models if models else RegionalDataMartModelEnum.get_enum_data().keys()
        self.regional_data_mart_models: List[ModelEnumValue] = [
            RegionalDataMartModelEnum.get_model_enum_value(model) for model in models
        ]

        self.logs_period_started_at = logs_period_started_at
        self.logs_period_ended_at = logs_period_ended_at
        self.logs_sub_period_days = logs_sub_period_days

        self.batch_size = batch_size

        self.get_logs_periods_sql = (
            """
            select min(created),
                   max(created),
                   row_batched
            from (
                select row_num,
                       ((row_num - 1) / {batch_size}) + 1 AS row_batched,
                       created
                from (
                    select row_number() over (order by created) as row_num,
                           created
                    from (
                        select *
                        from (
                        {ordered_rows}
                        ) as union_rows
                        order by created
                    ) as ordered
                ) as numbered
            ) as batched
            group by row_batched
            order by row_batched;
            """
        )

        self.ordered_rows_query = (
            """
            select created
            from {table_name}
            where created between '{period_started_at}' and '{period_ended_at}'
            """
        )

    def generate(self) -> List:
        """Генерирует список данных для формирования команд для сбора данных РВД."""
        params_for_commands = []

        for rdm_model in self.regional_data_mart_models:
            # Если подпериод указан, то формируется список с разбиением по logs_sub_period_days
            if self.logs_sub_period_days:
                # Получаем интервалы учебных годов
                academic_year_intervals = split_by_academic_years(
                    self.logs_period_started_at, self.logs_period_ended_at
                )

                for academic_year_start_datetime, academic_year_end_datetime in academic_year_intervals:
                    intervals_by_delta = split_interval_by_delta(
                        academic_year_start_datetime, academic_year_end_datetime, self.logs_sub_period_days
                    )
                    for start_datetime, end_datetime in intervals_by_delta:
                        params_for_commands.append(
                            {
                                'period_started_at': start_datetime,
                                'period_ended_at': end_datetime,
                                'model': rdm_model.key,
                                'logs_sub_period_days': self.logs_sub_period_days,
                            }
                        )

            # Если подпериод не указан, то формируется список с разбиением по batch_size
            else:
                ordered_rows_queries = [
                    self.ordered_rows_query.format(
                        table_name=model._meta.db_table,
                        period_started_at=self.logs_period_started_at,
                        period_ended_at=self.logs_period_ended_at,
                    )
                    for model in rdm_model.creating_trigger_models
                ]

                if hasattr(rdm_model, 'plugins_info'):
                    for app_name, app_models in rdm_model.plugins_info.items():
                        if apps.is_installed(app_name):
                            for app_model in app_models:
                                model = apps.get_model(app_model)
                                if model:
                                    ordered_rows_queries.append(
                                        self.ordered_rows_query.format(
                                            table_name=model._meta.db_table,
                                            period_started_at=self.logs_period_started_at.strftime(DATE_FORMAT),
                                            period_ended_at=self.logs_period_ended_at.strftime(DATE_FORMAT),
                                        )
                                    )

                ordered_rows_queries_sql = 'union'.join(ordered_rows_queries)

                temp_get_logs_periods_sql = self.get_logs_periods_sql.format(
                    batch_size=self.batch_size,
                    ordered_rows=ordered_rows_queries_sql,
                )

                with connection.cursor() as cursor:
                    cursor.execute(temp_get_logs_periods_sql)
                    rows = cursor.fetchall()

                for period_started_at, period_ended_at, batch_number in rows:
                    params_for_commands.append(
                        {
                            'period_started_at': period_started_at,
                            'period_ended_at': period_ended_at,
                            'model': rdm_model.key,
                            'logs_sub_period_days': self.logs_sub_period_days,
                        }
                    )

        return params_for_commands



from abc import (
    ABCMeta,
)

from function_tools.results import (
    BaseRunnableResult,
)


class WebEduRunnerResult(BaseRunnableResult, metaclass=ABCMeta):
    """
    Базовый класс результатов работы ранеров функций продуктов Образования.
    """


class WebEduFunctionResult(BaseRunnableResult, metaclass=ABCMeta):
    """
    Базовый класс результатов функций продуктов Образования.
    """



from abc import (
    ABCMeta,
)

from educommon import (
    logger,
)
from educommon.integration_entities.mixins import (
    EntitiesMixin,
)

from edu_rdm_integration.adapters.functions import (
    WebEduLazySavingPredefinedQueueGlobalHelperFunction,
)
from edu_rdm_integration.consts import (
    LOGS_DELIMITER,
)
from edu_rdm_integration.models import (
    CollectingDataSubStageStatus,
    CollectingExportedDataStage,
    CollectingExportedDataSubStage,
)


class BaseCollectingCalculatedDataFunction(
    EntitiesMixin,
    WebEduLazySavingPredefinedQueueGlobalHelperFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс функций сбора данных для интеграции с "Региональная витрина данных".
    """

    def __init__(self, *args, stage: CollectingExportedDataStage, **kwargs):
        super().__init__(*args, stage=stage, **kwargs)

        previous_sub_stage = CollectingExportedDataSubStage.objects.filter(
            function_id=self.uuid,
        ).order_by('started_at').only('pk').first()

        self._sub_stage = CollectingExportedDataSubStage.objects.create(
            stage=stage,
            function_id=self.uuid,
            previous_id=getattr(previous_sub_stage, 'pk', None),
        )

        logger.info(f'{LOGS_DELIMITER * 3}created {repr(self._sub_stage)}')

    def _before_prepare(self, *args, **kwargs):
        """
        Выполнение действий функций системы.
        """
        self._sub_stage.status_id = CollectingDataSubStageStatus.IN_PROGRESS.key
        self._sub_stage.save()

        logger.info(f'{LOGS_DELIMITER * 3}change status {repr(self._sub_stage)}')

    def run(self, *args, **kwargs):
        """
        Выполнение действий функции с дальнейшим сохранением объектов в базу при отсутствии ошибок.
        """
        super().run(*args, **kwargs)

        if self.result.errors:
            self._sub_stage.status_id = CollectingDataSubStageStatus.FAILED.key
        else:
            self._sub_stage.status_id = CollectingDataSubStageStatus.READY_TO_EXPORT.key

        self._sub_stage.save()

        logger.info(f'{LOGS_DELIMITER * 3}change status {repr(self._sub_stage)}')

        logger.info(f'{LOGS_DELIMITER * 3}{self.__class__.__name__} finished.')



from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
)

from django.db.models import (
    F,
    Max,
)
from django.utils import (
    timezone,
)
from m3.db import (
    BaseEnumerate,
)

from educommon import (
    logger,
)
from educommon.audit_log.models import (
    AuditLog,
)
from educommon.utils.date import (
    get_today_min_datetime,
)
from function_tools.managers import (
    RunnerManager,
)
from m3_db_utils.consts import (
    DEFAULT_ORDER_NUMBER,
)
from m3_db_utils.models import (
    ModelEnumValue,
)

from edu_rdm_integration.app_settings import (
    RDM_TRANSFER_TASK_TIMEDELTA,
)
from edu_rdm_integration.consts import (
    LOGS_SUB_PERIOD_DAYS,
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
)
from edu_rdm_integration.models import (
    CollectingDataSubStageStatus,
    CollectingExportedDataSubStage,
    RegionalDataMartModelEnum,
)
from edu_rdm_integration.signals import (
    manager_created,
)
from edu_rdm_integration.storages import (
    RegionalDataMartEntityStorage,
)


class BaseCollectModelsData:
    """Базовый класс сбора данных моделей РВД."""

    def __init__(
        self,
        models: Iterable[str],
        logs_period_started_at=datetime.combine(date.today(), time.min),
        logs_period_ended_at=datetime.combine(date.today(), time.min),
        **kwargs,
    ):
        # Если модели не указаны, берется значение по умолчанию - все модели:
        models = models if models else RegionalDataMartModelEnum.get_enum_data().keys()
        self.models: List[ModelEnumValue] = [
            RegionalDataMartModelEnum.get_model_enum_value(model) for model in models
        ]

        self.logs_period_started_at = logs_period_started_at
        self.logs_period_ended_at = logs_period_ended_at

        # Классы менеджеров Функций, которые должны быть запущены для сбора данных моделей РВД
        self._collecting_data_managers: Dict[str, Type[RunnerManager]] = {}

        # Результаты работы Функций сбора данных моделей РВД
        self._collecting_data_results = []

        # Идентификатор CollectDataCommandProgress для передачи сигналу manager_created
        self.command_id: Optional[int] = kwargs.get('command_id')

    def _find_collecting_models_data_managers(self):
        """
        Поиск менеджеров Функций, которые должны быть запущены для сбора данных моделей РВД.
        """
        logger.info('collecting_models_data_managers..')

        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        collecting_models_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA},
        )

        for model_enum in self.models:
            manager_class = collecting_models_data_managers_map.get(model_enum.key)

            if manager_class:
                self._collecting_data_managers[model_enum.key] = manager_class

        logger.info('collecting models data managers finished')

    def _collect_models_data(self, *args, logs: Optional[Dict[str, List[AuditLog]]] = None, **kwargs):
        """
        Запуск Функций по формированию данных моделей РВД из логов.
        """
        logger.info('collect models data..')

        kwargs['logs_period_started_at'] = self.logs_period_started_at
        kwargs['logs_period_ended_at'] = self.logs_period_ended_at

        for model_key, manager_class in self._collecting_data_managers.items():
            model_logs = logs.get(model_key) if logs else None
            manager = manager_class(*args, logs=model_logs, **kwargs)

            if self.command_id:
                # Подается сигнал, что менеджер создан:
                manager_created.send(sender=manager,
                                     command_id=self.command_id)

            manager.run()

            self._collecting_data_results.append(manager.result)

        logger.info('collecting entities data finished.')

    def collect(self):
        """Запускает сбор данных."""
        logger.info(
            f'start collecting data of models - {", ".join([model.key for model in self.models])}..'
        )

        self._find_collecting_models_data_managers()
        self._collect_models_data()

        logger.info('collecting models data finished.')


class BaseCollectModelsDataByGeneratingLogs(BaseCollectModelsData):
    """Сбор данных моделей РВД на основе существующих в БД данных моделей ЭШ.

    Можно регулировать, для каких моделей должен быть произведен сбор данных, и период, за который должны
    быть собраны логи. Логи формируются в процессе выполнения команды при помощи генератора логов
    EduSchoolLogGenerator для указанной модели.
    """

    def __init__(
        self,
        models: Iterable[str],
        logs_period_started_at=datetime.combine(date.today(), time.min),
        logs_period_ended_at=datetime.combine(date.today(), time.min),
        logs_sub_period_days=LOGS_SUB_PERIOD_DAYS,
        school_ids=(),
        **kwargs,
    ):
        super().__init__(models, logs_period_started_at, logs_period_ended_at, **kwargs)

        self.logs_sub_period_days = logs_sub_period_days
        # Школы, для которых производится выгрузка
        self.school_ids = school_ids
        # Генератор логов
        self.log_generator = self._prepare_log_generator()

    def _prepare_log_generator(self) -> 'BaseEduLogGenerator':
        """Возвращает генератор логов."""
        raise NotImplementedError

    def _generate_logs_by_subperiod(self):
        """
        Генерация логов с учетом подпериодов.
        """
        temp_logs_period_started_at = self.logs_period_started_at
        temp_logs_period_ended_at = self.logs_period_started_at + timedelta(days=self.logs_sub_period_days)

        if temp_logs_period_ended_at > self.logs_period_ended_at:
            temp_logs_period_ended_at = self.logs_period_ended_at

        temp_logs: Dict[str, List[AuditLog]] = {}

        while temp_logs_period_started_at < temp_logs_period_ended_at <= self.logs_period_ended_at:
            for model in self.models:
                logs = self.log_generator.generate(
                    model=model,
                    logs_period_started_at=temp_logs_period_started_at,
                    logs_period_ended_at=temp_logs_period_ended_at,
                    school_ids=self.school_ids,
                )

                temp_logs[model.key] = logs

            yield temp_logs, temp_logs_period_started_at, temp_logs_period_ended_at

            temp_logs_period_started_at = temp_logs_period_ended_at
            temp_logs_period_ended_at += timedelta(days=self.logs_sub_period_days)

            if temp_logs_period_ended_at > self.logs_period_ended_at:
                temp_logs_period_ended_at = self.logs_period_ended_at

            temp_logs.clear()

    def _generate_logs_for_all_period(self):
        """
        Генерация логов за весь период.
        """
        temp_logs: Dict[str, List[AuditLog]] = {}

        for model in self.models:
            logs = self.log_generator.generate(
                model=model,
                logs_period_started_at=self.logs_period_started_at,
                logs_period_ended_at=self.logs_period_ended_at,
                school_ids=self.school_ids,
            )

            temp_logs[model.key] = logs

        return [(temp_logs, self.logs_period_started_at, self.logs_period_ended_at)]

    def _generate_logs(self) -> List[Tuple[Dict[str, List[AuditLog]], datetime, datetime]]:
        """
        Генерация логов.

        Осуществляет генерацию логов по уже существующим записям в базе данных. В качестве параметров указываются
        начало и конец периода сбора логов, размер подпериодов, на которые должен быть разбит основной период.
        Генерация логов производится только для указанных моделей.
        """
        if self.logs_sub_period_days:
            logs = self._generate_logs_by_subperiod()
        else:
            logs = self._generate_logs_for_all_period()

        return logs

    def collect(self):
        """Главный метод, через который происходит сбор данных."""
        logger.info(
            f'start collecting data of models - {", ".join([model.key for model in self.models])}..'
        )
        self._find_collecting_models_data_managers()

        temp_kwargs = {}

        for logs, logs_period_started_at, logs_period_ended_at in self._generate_logs():
            temp_kwargs['logs_period_started_at'] = logs_period_started_at
            temp_kwargs['logs_period_ended_at'] = logs_period_ended_at

            self._collect_models_data(logs=logs, **temp_kwargs)

        logger.info('collecting models data finished.')


class BaseCollectLatestModelsData(BaseCollectModelsData):
    """
    Сбор данных моделей РВД на основе логов за период с последней сборки до указанной даты.
    """

    def __init__(
        self,
        models: Iterable[str],
        logs_period_started_at=datetime.combine(date.today(), time.min),
        logs_period_ended_at=datetime.combine(date.today(), time.min),
        **kwargs,
    ):
        super().__init__(models, logs_period_started_at, logs_period_ended_at, **kwargs)

        # Перечисление с очередностью сбора данных по моделям
        self.order_number_model_enum = self._get_order_number_model_enum()

    def _get_order_number_model_enum(self) -> BaseEnumerate:
        """Возвращает перечисление с очередностью сбора данных по моделям."""
        raise NotImplementedError

    def _find_collecting_models_data_managers(self, *args, **kwargs) -> None:
        """
        Поиск менеджеров Функций, которые должны быть запущены для сбора данных моделей РВД.
        """
        logger.info('collecting_models_data_managers..')

        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        collecting_models_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA},
        )

        export_models = filter(
            lambda entity: entity.order_number not in (None, DEFAULT_ORDER_NUMBER), self.models
        )

        # Сортируем модели по очередности загрузки
        sorted_models = sorted(
            export_models,
            key=lambda model: self.order_number_model_enum.get_constant_value_by_name(model.key)
        )

        for model_enum in sorted_models:
            manager_class = collecting_models_data_managers_map.get(model_enum.key)

            if manager_class:
                self._collecting_data_managers[model_enum.key] = manager_class

        logger.info('collecting models data managers finished')

    def _get_last_finished_entity_upload(self) -> Dict[str, datetime]:
        """
        Возвращает словарь с uuid менеджера и датой последней успешной выгрузки по указанным сущностям.
        """
        manager_to_last_date = CollectingExportedDataSubStage.objects.annotate(
            max_logs_period_ended_at=Max('stage__logs_period_ended_at'),
        ).values(
            'stage__manager_id', 'max_logs_period_ended_at',
        ).annotate(
            manager_id=F('stage__manager_id'),
            date_end=F('max_logs_period_ended_at'),
        ).filter(
            status_id=CollectingDataSubStageStatus.READY_TO_EXPORT.key,
            manager_id__in=[m.uuid for m in self._collecting_data_managers.values()],
        ).values('manager_id', 'date_end')

        return {str(m['manager_id']): m['date_end'] for m in manager_to_last_date}

    def _collect_models_data(self, *args, logs: Optional[Dict[str, List[AuditLog]]] = None, **kwargs) -> None:
        """
        Запуск Функций по формированию данных из логов для дальнейшей выгрузки.
        """
        logger.info('collect entities data..')

        last_finished_entity_upload = self._get_last_finished_entity_upload()

        if logs is None:
            logs = {}

        for entity_key, manager_class in self._collecting_data_managers.items():
            entity_logs = logs.get(entity_key)

            kwargs['logs_period_started_at'] = (
                last_finished_entity_upload.get(manager_class.uuid)
                or get_today_min_datetime()
            )
            kwargs['logs_period_ended_at'] = timezone.now()

            if kwargs['logs_period_started_at'] > kwargs['logs_period_ended_at']:
                kwargs['logs_period_started_at'] = kwargs['logs_period_ended_at'] - datetime.timedelta(
                    seconds=RDM_TRANSFER_TASK_TIMEDELTA
                )

            manager = manager_class(*args, logs=entity_logs, **kwargs)

            if self.command_id:
                # Подается сигнал, что менеджер создан:
                manager_created.send(sender=manager,
                                     command_id=self.command_id)

            manager.run()

            self._collecting_data_results.append(manager.result)

        logger.info('collecting entities data finished.')


class CollectModelsData(BaseCollectModelsData):
    """Сбор данных моделей РВД за указанных период по существующим логам."""



from edu_rdm_integration.collect_data.collect import (
    BaseCollectModelsData,
    CollectModelsData,
)
from edu_rdm_integration.management.general import (
    BaseCollectModelDataCommand,
)


class Command(BaseCollectModelDataCommand):
    """
    Команда для сбора данных моделей РВД за указанных период по существующим логам.
    """

    # flake8: noqa: A003
    help = 'Команда для сбора данных моделей РВД за указанных период по существующим логам.'

    def _prepare_collect_models_data_class(self, *args, **kwargs) -> BaseCollectModelsData:
        """Возвращает объект класса сбора данных моделей РВД."""
        return CollectModelsData(
            models=kwargs.get('models'),
            logs_period_started_at=kwargs.get('logs_period_started_at'),
            logs_period_ended_at=kwargs.get('logs_period_ended_at'),
        )



import os
from datetime import (
    datetime,
    time,
    timedelta,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
)

from django.conf import (
    settings,
)

from edu_rdm_integration.apps import (
    EduRDMIntegrationConfig,
)
from edu_rdm_integration.consts import (
    ACADEMIC_YEAR,
)


if TYPE_CHECKING:
    from edu_rdm_integration.models import (
        BaseEntityModel,
    )


def get_exporting_data_stage_attachment_path(instance, filename):
    """Возвращает путь до файла-вложения в этап выгрузки данных сущности.

    Args:
        instance: объект ExportingDataStage
        filename: имя загружаемого файла

    Returns:
        Строковое представление пути
    """
    datetime_now = datetime.now()

    return os.path.join(
        settings.UPLOADS,
        EduRDMIntegrationConfig.label,
        datetime_now.strftime('%Y/%m/%d'),
        instance.exporting_data_sub_stage.__class__.__name__.lower(),
        str(instance.exporting_data_sub_stage_id),
        str(instance.operation),
        filename,
    )


def update_fields(
    entity: 'BaseEntityModel',
    field_values: Dict[str, Any],
    mapping: Dict[str, str]
) -> None:
    """Обновление значений полей сущности по измененным полям модели.

    :param entity: Выгружаемая сущность
    :param field_values: Словарь с измененными данными модели
    :param mapping: Словарь маппинга полей модели к полям сущности
    """
    for model_field, entity_field in mapping.items():
        if model_field in field_values:
            setattr(entity, entity_field, field_values[model_field])


def get_isoformat_timezone():
    """Возвращает временную зонну в ISO представлении."""
    return datetime.now().astimezone().isoformat()[-6:]


def split_by_academic_years(start_date: datetime, end_date: datetime) -> List[tuple]:
    """Разбивает исходный интервал из команды по учебным годам."""
    academic_year_end_date = datetime(
        start_date.year, ACADEMIC_YEAR['end_month'], ACADEMIC_YEAR['end_day'],

    )
    if academic_year_end_date <= start_date:
        academic_year_end_date = datetime(
            academic_year_end_date.year + 1,
            ACADEMIC_YEAR['end_month'],
            ACADEMIC_YEAR['end_day'],
            time.max.hour,
            time.max.minute,
            time.max.second,
        )

    splitted_interval = []
    splitted_interval.append(start_date)

    while start_date < academic_year_end_date < end_date:
        splitted_interval.append(academic_year_end_date)

        academic_year_end_date = datetime(
            academic_year_end_date.year + 1,
            ACADEMIC_YEAR['end_month'],
            ACADEMIC_YEAR['end_day'],
            time.max.hour,
            time.max.minute,
            time.max.second,
        )
        academic_year_start_date = datetime(
            academic_year_end_date.year - 1,
            ACADEMIC_YEAR['start_month'],
            ACADEMIC_YEAR['start_day'],
            time.min.hour,
            time.min.minute,
            time.min.second,
        )

        splitted_interval.append(academic_year_start_date)

    splitted_interval.append(end_date)

    iter_interval = iter(splitted_interval)
    intervals = [*zip(iter_interval, iter_interval)]

    return intervals


def split_interval_by_delta(start_date: datetime, end_date: datetime, days_delta: int) -> List[tuple]:
    """Разбивает учебный год на интервалы по дельте."""
    subinterval_end = start_date + timedelta(days=days_delta)

    splitted_interval = []
    splitted_interval.append(start_date)

    # В данном цикле проверять нужно именно даты чтобы не получилась ситуация что две одинаковые даты
    # попадут в интервалы и различаться будут лишь временем (минимальная дельта 1 день т.е. 24 часа)
    while start_date.date() < subinterval_end.date() < end_date.date():
        splitted_interval.append(subinterval_end)
        subinterval_start = subinterval_end + timedelta(days=1)
        splitted_interval.append(subinterval_start)
        subinterval_end = subinterval_end + timedelta(days=days_delta + 1)

    splitted_interval.append(end_date)

    iter_interval = iter(splitted_interval)
    intervals = [*zip(iter_interval, iter_interval)]

    return intervals



from typing import (
    Type,
)

from edu_rdm_integration.collect_data.base.runners import (
    BaseCollectingDataRunner,
)
from edu_rdm_integration.collect_data.non_calculated.base.helpers import (
    BaseCollectingExportedDataRunnerHelper,
)
from edu_rdm_integration.collect_data.non_calculated.base.results import (
    BaseCollectingExportedDataRunnerResult,
)
from edu_rdm_integration.collect_data.non_calculated.base.validators import (
    BaseCollectingExportedDataRunnerValidator,
)


class BaseCollectingExportedDataRunner(BaseCollectingDataRunner):
    """
    Базовый класс ранеров функций сбора данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_helper_class(self) -> Type[BaseCollectingExportedDataRunnerHelper]:
        """
        Возвращает класс помощника ранера функции.
        """
        return BaseCollectingExportedDataRunnerHelper

    def _prepare_validator_class(self) -> Type[BaseCollectingExportedDataRunnerValidator]:
        """
        Возвращает класс валидатора ранера функции.
        """
        return BaseCollectingExportedDataRunnerValidator

    def _prepare_result_class(self) -> Type[BaseCollectingExportedDataRunnerResult]:
        """
        Возвращает класс результата ранера функции.
        """
        return BaseCollectingExportedDataRunnerResult



from abc import (
    abstractmethod,
)
from collections import (
    defaultdict,
)
from operator import (
    itemgetter,
)
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Union,
)

from educommon.utils.conversion import (
    int_or_none,
)

from edu_rdm_integration.adapters.caches import (
    WebEduEntityCache,
    WebEduFunctionCacheStorage,
    WebEduRunnerCacheStorage,
)
from edu_rdm_integration.collect_data.calculated.base.consts import (
    LOOKUP_SEP,
)


class BaseCollectingCalculatedExportedDataRunnerCacheStorage(WebEduRunnerCacheStorage):
    """
    Базовый кеш помощников ранеров функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """


class BaseCollectingCalculatedExportedDataFunctionCacheStorage(WebEduFunctionCacheStorage):
    """
    Базовый кеш помощников функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """

    def __init__(self, raw_logs, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Необработанные логи как есть.
        self.raw_logs = raw_logs

        # Подготовленные логи в виде:
        # {
        #     'person.person': {
        #         48939: [(1, {'surname': 'Иванов', 'snils': '157-283-394 92'...}), (2, {...})],
        #         44281: [(2, {...}), (2, {...}), ...],
        #         12600: [(3, {...})],
        #     },
        #     'schoolchild.schoolchild': {
        #         ...
        #     },
        # }
        self.logs = defaultdict(lambda: defaultdict(list))

    def _reformat_logs(self):
        """
        Производится преобразование логов к удобному для работы виду.

        Предполагается вложенные словари. На первом уровне ключом будет название
        модели, на втором идентификатор записи.
        """

    def _prepare_logs(self):
        """
        Подготовка логов для дальнейшей работы.
        """
        self._reformat_logs()

    @staticmethod
    def _add_id_to_set(ids: set, object_id: Union[int, str, None]) -> None:
        """Добавление значения id во множество ids."""
        object_id = int_or_none(object_id)
        if object_id:
            ids.add(object_id)

    @abstractmethod
    def _collect_product_model_ids(self):
        """Собирает идентификаторы записей моделей продукта с упором на логи."""

    @abstractmethod
    def _prepare_caches(self):
        """Формирование кэшей."""

    def _prepare(self, *args, **kwargs):
        """Запускает формирование кэша помощника Функции."""
        self._prepare_logs()
        self._collect_product_model_ids()
        self._prepare_caches()


class WebEduEntityValueCache(WebEduEntityCache):
    """Расширение базового класс для возможности получать значения из QuerySet."""

    def __init__(self, *args, values, **kwargs):
        self._values = values
        self._filters = {}
        self._filtered_entity_cache = []
        self.is_filtered = False

        self.filter_functions = {
            'range': lambda r: lambda v: r[0] <= v <= r[1],
            'in': lambda r: lambda v: v in r,
            'eq': lambda r: lambda v: v == r
        }

        super().__init__(*args, **kwargs)

    def _prepare_entities(self):
        """
        Получение выборки объектов модели по указанными параметрам.
        """
        self._entities = self._actual_entities_queryset.filter(
            **self._additional_filter_params,
        )

        if self._only_fields:
            self._entities = self._entities.only(*self._only_fields)

        if self._values:
            self._entities = self._entities.values(*self._values)

        self._entities = self._entities.distinct()

    def filter(self, **kwargs):  # noqa: A003
        """
        Установка фильтра для выборки данных.

        При повторном применении фильтра, ранее установленные параметры сбрасываются.

        Пример использования:
        some_objects_list = cache.filter(
            id__range=(1,10), elements_id__in=[1,2,3]
        ).values_list('id')
        some_objects_list => [1,5,6]

        some_objects_list = cache.filter(
            id__range=(1,10), elements_id__in=[1,2,3]
        ).values_list('id', 'elements_id')
        some_objects_list => [(1,1), (5,2), (6,3)]
        """
        self._clear_filter()

        for attr, value_filter in kwargs.items():
            lookup_attr = attr.split(LOOKUP_SEP)
            if lookup_attr[-1] in self.filter_functions:
                attr = LOOKUP_SEP.join(lookup_attr[:-1])
                lookup_value = self.filter_functions.get(lookup_attr[-1])(value_filter)
            else:
                lookup_value = self.filter_functions.get('eq')(value_filter)

            self._filters[attr] = lookup_value

        return self

    def _filtered_entities(self, entities, filters):
        """Фильтрует сущности на основании применяемых фильтров."""
        self._filtered_entity_cache = []
        if filters:
            for entity in entities:
                if all(
                    func_filter(entity.get(attr)) for attr, func_filter in filters.items()
                ):
                    self._filtered_entity_cache.append(entity)
        else:
            self._filtered_entity_cache = entities

    def _get_filter(self) -> Dict[str, Callable[..., Any]]:
        """Возвращает наименование атрибута и фильтр по нему."""
        return self._filters

    def _clear_filter(self):
        """Сброс примененных фильтров и очистка кеша."""
        self.is_filtered = False
        self._filtered_entity_cache = []
        self._filters = {}

    def values_list(self, *args, **kwargs) -> Optional[List[Any]]:
        """
        Получение списка кортежей или значений согласно заданным параметрам фильтрации.

        Пример использования:
        some_objects_list = cache.filter(
            id__range=(1,10), elements_id__in=[1,2,3]
        )
        some_objects_list = some_objects_list.values_list('id', 'elements_id')
        some_objects_list => [(1,1), (5,2), (6,3)]
        """
        if not self.is_filtered:
            self._filtered_entities(self._entities, self._get_filter())
            self.is_filtered = True

        fields_getter = itemgetter(*args)

        return [fields_getter(entity) for entity in self._filtered_entity_cache]

    def exists(self) -> bool:
        """
        Проверка существования отфильтрованных данных.

        Кеширует результат фильтрации данных, для исключения повторного прохода.
        """
        if not self.is_filtered:
            self._filtered_entities(self._entities, self._get_filter())
            self.is_filtered = True

        return bool(self._filtered_entity_cache)



from edu_rdm_integration.adapters.results import (
    WebEduFunctionResult,
    WebEduRunnerResult,
)


class BaseCollectingCalculatedExportedDataRunnerResult(WebEduRunnerResult):
    """
    Базовый класс результатов работы ранеров функций сбора расчетных данных для "Региональной витрины данных".
    """


class BaseCollectingCalculatedExportedDataFunctionResult(WebEduFunctionResult):
    """
    Базовый класс результатов работы функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """
